"""
TODO: The plugin already gets imported and used, even if it isn't in __all__ here!
Is it ready to be used though? Is there a way we could allow an "opt-in" mechanism for the
plugin to be used instead?
"""
# from .auto_schema_plugin import AutoSchemaPlugin, AutoSchemaPluginConfig

__all__ = [
    # "AutoSchemaPlugin",
    # "AutoSchemaPluginConfig",
]



"""Scripts that creates Schemas for hydra config files.

This is very helpful when using Hydra! It shows the user what options are available, along with their
description and default values, and displays errors if you have config files with invalid values.

## todos
- [ ] Support '???' as a value for any property.
- [ ] Modify the schema to support omegaconf directives like ${oc.env:VAR_NAME} and our custom directives like ${instance_attr} and so on.
- [ ] todo: Make a hydra plugin that creates the schemas for configs when hydra is loading stuff.
"""

from __future__ import annotations

import copy
import dataclasses
import datetime
import inspect
import json
import os.path
import subprocess
import sys
import typing
import warnings
from logging import getLogger as get_logger
from pathlib import Path
from typing import Any, TypeVar

import docstring_parser as dp
import hydra.conf
import hydra.errors
import hydra.plugins
import hydra.utils
import hydra_zen
import omegaconf
import pydantic
import pydantic.schema
import tqdm
from hydra._internal.config_loader_impl import ConfigLoaderImpl
from hydra._internal.config_search_path_impl import ConfigSearchPathImpl
from hydra.core.config_search_path import ConfigSearchPath
from hydra.core.config_store import ConfigNode, ConfigStore
from hydra.core.plugins import Plugins
from hydra.plugins.search_path_plugin import SearchPathPlugin
from hydra.types import RunMode
from omegaconf import DictConfig, OmegaConf
from pydantic.json_schema import GenerateJsonSchema, JsonSchemaValue
from pydantic_core import core_schema
from tqdm.rich import tqdm_rich

from hydra_auto_schema.customize import (
    custom_enum_schemas,
    custom_hydra_zen_builds_args,
    schema_conflict_handlers,
)
from hydra_auto_schema.hydra_schema import (
    HYDRA_CONFIG_SCHEMA,
    ObjectSchema,
    PropertySchema,
    Schema,
)
from hydra_auto_schema.utils import merge_dicts, pretty_path

logger = get_logger(__name__)

config = None

K = TypeVar("K")
V = TypeVar("V")
PossiblyNestedDict = dict[K, V | "PossiblyNestedDict[K, V]"]


def _yaml_files_in(configs_dir: Path) -> list[Path]:
    # Ignores .venv subfiles if the `configs_dir` isn't itself in a ".venv" directory.
    if ".venv" in configs_dir.parts:
        return list(configs_dir.rglob("*.yaml")) + list(configs_dir.rglob("*.yml"))
    return list(
        p for p in configs_dir.rglob("*.yaml") if ".venv" not in p.parts
    ) + list(p for p in configs_dir.rglob("*.yml") if ".venv" not in p.parts)


def add_schemas_to_all_hydra_configs(
    repo_root: Path,
    configs_dir: Path,
    schemas_dir: Path | None = None,
    regen_schemas: bool = False,
    stop_on_error: bool = False,
    quiet: bool = False,
    add_headers: bool | None = False,
    config_store: ConfigStore | None = None,
):
    """Adds schemas to all the passed Hydra config files.

    Parameters:
        repo_root: The root directory of the repository.
        configs_dir: The directory containing the Hydra config files.
        schemas_dir: The directory to store the generated schema files. Defaults to ".schemas" in the repo_root.
        regen_schemas: If True, regenerate schemas even if they already exist. Defaults to False.
        stop_on_error: If True, raise an exception on error. Defaults to False.
        quiet: If True, suppress progress bar output. Defaults to False.
        add_headers: Determines how to associate schema files with config files.

            - If None, try adding to VSCode settings first, then fallback to adding headers.
            - If False, only use VSCode settings.
            - If True, only add headers.
    """
    config_store = config_store or ConfigStore.instance()
    config_files = _yaml_files_in(configs_dir)
    if not config_files:
        if stop_on_error:
            raise RuntimeError("No config files were found!")
        else:
            warnings.warn(RuntimeWarning("No config files were found! Skipping."))
        # return

    if schemas_dir is None:
        schemas_dir = repo_root / ".schemas"

    if schemas_dir.is_relative_to(repo_root):
        _add_schemas_dir_to_gitignore(schemas_dir, repo_root=repo_root)

    config_file_to_schema_file: dict[Path, Path] = {}
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=tqdm.TqdmExperimentalWarning)
        pbar = tqdm_rich(
            config_files,
            desc="Creating schemas for Hydra config files...",
            total=len(config_files),
            leave=False,
            disable=quiet,
        )

    for config_file in pbar:
        pretty_config_file_name = config_file.relative_to(configs_dir)
        schema_file = get_schema_file_path(config_file, schemas_dir)

        # Check the modification time. If the config file was modified after the schema file,
        # regen the schema file.

        if schema_file.exists():
            schema_file_modified_time = datetime.datetime.fromtimestamp(
                schema_file.stat().st_mtime
            )
            config_file_modified_time = datetime.datetime.fromtimestamp(
                config_file.stat().st_mtime
            )
            # Add a delay to account for the time it takes to modify the the config file at the end
            # to remove a header.
            if (
                config_file_modified_time - schema_file_modified_time
            ) > datetime.timedelta(seconds=10):
                logger.info(
                    f"Config file {pretty_config_file_name} was modified, regenerating the schema."
                )
            elif regen_schemas:
                pass  # regenerate it.
            elif sys.platform == "linux" and _is_incomplete_schema(schema_file):
                logger.info(
                    f"Unable to properly create the schema for {pretty_config_file_name} last time. Trying again."
                )
            else:
                logger.debug(
                    f"Schema file {pretty_path(schema_file)} was already successfully created. Skipping."
                )
                continue

        pbar.set_postfix_str(f"Creating schema for {pretty_config_file_name}")

        # We'll modify the config store so that we treat structured configs as if they
        # had a _target_ corresponding to the structured class.
        # This helps us create the schemas.
        # We later reset this to not affect the config loading in the Hydra application.
        from hydra.core.singleton import Singleton  # noqa

        state_backup = Singleton.get_state()

        try:
            logger.debug(f"Creating a schema for {pretty_config_file_name}")

            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=UserWarning)
                # TODO: Can we somehow get that the ConfigStore entries should
                # be used as targets?
                #     Singleton._instances.pop(ConfigStore)
                # else:
                # Maybe using the _convert_ param of Hydra?
                config = load_config(
                    config_file,
                    configs_dir=configs_dir,
                    repo_root=repo_root,
                    config_store=config_store,
                )
            schema = _create_schema_for_config(
                config,
                config_file=config_file,
                configs_dir=configs_dir,
                repo_root=repo_root,
                config_store=config_store,
            )
            schema_file.parent.mkdir(exist_ok=True, parents=True)
            schema_file.write_text(json.dumps(schema, indent=2).rstrip() + "\n\n")
            if sys.platform == "linux":
                _set_is_incomplete_schema(schema_file, False)
        except (
            pydantic.errors.PydanticSchemaGenerationError,
            hydra.errors.MissingConfigException,
            hydra.errors.ConfigCompositionException,
            omegaconf.errors.InterpolationResolutionError,
            Exception,  # todo: remove this to harden the code.
        ) as exc:
            logger.warning(
                RuntimeWarning(
                    f"Unable to create a schema for config {pretty_config_file_name}: {exc}"
                )
            )
            if stop_on_error:
                raise

            schema = copy.deepcopy(HYDRA_CONFIG_SCHEMA)
            schema["additionalProperties"] = True
            schema["title"] = f"Partial schema for {pretty_config_file_name}"
            schema[
                "description"
            ] = f"(errors occurred while trying to create the schema from the signature:\n{exc}"
            schemas_dir.mkdir(exist_ok=True, parents=True)
            schema_file.write_text(json.dumps(schema, indent=2) + "\n")
            if sys.platform == "linux":
                _set_is_incomplete_schema(schema_file, True)
        finally:
            # Reset the config store / search path / etc to what they were before.
            Singleton.set_state(state_backup)

        config_file_to_schema_file[config_file] = schema_file

    # Option 1: Add a vscode setting that associates the schema file with the yaml files. (less intrusive perhaps).
    # Option 2: Add a header to the yaml files that points to the schema file.

    # If add_headers is None, try option 1, then fallback to option 2.
    # If add_headers is False, only use option 1
    # If add_headers is True, only use option 2

    if not add_headers:
        try:
            _install_yaml_vscode_extension()
        except OSError:
            pass

        try:
            _add_schemas_to_vscode_settings(
                config_file_to_schema_file, repo_root=repo_root
            )
        except Exception as exc:
            logger.error(
                f"Unable to write schemas in the vscode settings file. "
                f"Falling back to adding a header to config files. (exc={exc})"
            )
            if add_headers is not None:
                # Unable to do it. Don't try to add headers, just return.
                return
        else:
            # Success. Return.
            return

    logger.debug("Adding headers to config files to point to the schemas to use.")
    for config_file, schema_file in config_file_to_schema_file.items():
        _add_schema_header(config_file, schema_path=schema_file)


def _get_gitignore_path(repo_root: Path) -> Path:
    for parent in repo_root.parents:
        if (gitignore_file := (parent / ".gitignore")).exists():
            return gitignore_file
    return repo_root / ".gitignore"


def _add_schemas_dir_to_gitignore(schemas_dir: Path, repo_root: Path):
    gitignore_file = _get_gitignore_path(repo_root)
    if not schemas_dir.is_relative_to(gitignore_file.parent):
        # The schemas dir is not under the same directory as the gitignore file, so we can't add it.
        return

    _rel = schemas_dir.relative_to(gitignore_file.parent)
    if not gitignore_file.exists():
        gitignore_file.write_text(f"{_rel}\n")
        return
    if not any(
        line.startswith(str(_rel)) for line in gitignore_file.read_text().splitlines()
    ):
        logger.info(
            f"Adding entry in .gitignore for the schemas directory ({schemas_dir})"
        )
        with gitignore_file.open("a") as f:
            f.write(f"{_rel}\n")


_incomplete_schema_xattr = "user.schema_error"


def _is_incomplete_schema(schema_file: Path) -> bool:
    try:
        return os.getxattr(schema_file, _incomplete_schema_xattr) == bytes(1)
    except OSError:
        return False


def _set_is_incomplete_schema(schema_file: Path, val: bool):
    os.setxattr(schema_file, _incomplete_schema_xattr, bytes(val))


def _relative_to_cwd(p: str | Path):
    return Path(p).relative_to(Path.cwd())


def _install_yaml_vscode_extension():
    logger.debug(
        "Running `code --install-extension redhat.vscode-yaml` to install the yaml extension for vscode."
    )
    output = subprocess.check_output(
        ("code", "--install-extension", "redhat.vscode-yaml"), text=True
    )
    logger.debug(output)


def _read_json(file: Path) -> dict:
    file_contents = file.read_text()
    if not file_contents:
        return {}
    # Remove any trailing commas from the content:
    file_contents = (
        # Remove any trailing "," that would make it invalid JSON.
        "".join(file_contents.split()).replace(",}", "}").replace(",]", "]")
    )
    return json.loads(file_contents)


def _add_schemas_to_vscode_settings(
    config_file_to_schema_file: dict[Path, Path],
    repo_root: Path,
) -> None:
    # Make the vscode settings file if necessary:
    vscode_dir = repo_root / ".vscode"
    vscode_dir.mkdir(exist_ok=True, parents=False)
    vscode_settings_file = vscode_dir / "settings.json"
    vscode_settings_file.touch(exist_ok=True)

    # TODO: Use Jsonc to load the file and preserve comments?
    logger.debug(f"Reading the VsCode settings file at {vscode_settings_file}.")
    vscode_settings: dict[str, Any] = _read_json(vscode_settings_file)

    # logger.debug(f"Vscode settings: {vscode_settings}")
    # Avoid the popup and do users a favour by disabling telemetry.
    vscode_settings.setdefault("redhat.telemetry.enabled", False)

    # TODO: Should probably overwrite the schemas entry if we're passed the --regen-schemas flag,
    # since otherwise we might accumulate schemas for configs that aren't there anymore.
    yaml_schemas_setting: dict[str, str | list[str]] = vscode_settings.setdefault(
        "yaml.schemas", {}
    )

    # Write all the schemas
    for config_file, schema_file in config_file_to_schema_file.items():
        assert schema_file.exists()

        _root = vscode_settings_file.parent.parent
        schema_key = str(schema_file.relative_to(_root))
        path_to_add = str(config_file.absolute())

        if schema_key not in yaml_schemas_setting:
            yaml_schemas_setting[schema_key] = path_to_add
        elif isinstance(
            files_associated_with_schema := yaml_schemas_setting[schema_key], str
        ):
            files = sorted(set([files_associated_with_schema, path_to_add]))
            yaml_schemas_setting[schema_key] = files[0] if len(files) == 1 else files
        else:
            files = sorted(set(files_associated_with_schema + [path_to_add]))
            yaml_schemas_setting[schema_key] = files[0] if len(files) == 1 else files

    vscode_settings_file.write_text(json.dumps(vscode_settings, indent=2))
    logger.debug(
        f"Updated the yaml schemas in the vscode settings file at {vscode_settings_file}."
    )

    # If this worked, then remove any schema directives from the config files.
    for config_file, schema_file in config_file_to_schema_file.items():
        assert schema_file.exists()
        config_lines = config_file.read_text().splitlines()
        lines_to_remove: list[int] = []
        for i, line in enumerate(config_lines):
            if line.strip().startswith("# yaml-language-server: $schema="):
                lines_to_remove.append(i)
        for line_to_remove in reversed(lines_to_remove):
            config_lines.pop(line_to_remove)
        config_file.write_text(
            "\n".join(config_lines).rstrip() + ("\n" if config_lines else "")
        )


def get_schema_file_path(config_file: Path, schemas_dir: Path):
    config_group = config_file.parent
    schema_file = schemas_dir / f"{config_group.name}_{config_file.stem}_schema.json"
    return schema_file


def _all_subentries_with_target(config: dict) -> dict[tuple[str, ...], dict]:
    """Iterator that yields all the nested config entries that have a _target_."""
    entries = {}
    if "_target_" in config:
        entries[()] = config

    for key, value in config.items():
        if isinstance(value, dict):
            for subkey, subvalue in _all_subentries_with_target(value).items():
                entries[(key, *subkey)] = subvalue
    return entries


def _create_schema_for_config(
    config: dict | DictConfig,
    config_file: Path,
    configs_dir: Path,
    repo_root: Path,
    config_store: ConfigStore | None,
) -> Schema | ObjectSchema:
    """IDEA: Create a schema for the given config.

    - If you encounter a key, add it to the schema.
    - If you encounter a value with a _target_, use a dedicated function to get the schema for that target, and merge it into the current schema.
    - Only the top-level config (`config`) can have a `defaults: list[str]` key.
        - Should ideally load the defaults and merge this schema on top of them.
    """

    # Start from the base schema for any Hydra configs.
    schema = copy.deepcopy(HYDRA_CONFIG_SCHEMA)

    pretty_path = config_file.relative_to(configs_dir) if configs_dir else config_file
    schema["title"] = f"Auto-generated schema for {pretty_path}"

    # NOTE: Config files can also not have a target.
    # This sets additionalProperties to `false` if there is a structured config in the defaults
    # list, since `schema` will have been given a `_target_` above.
    schema["additionalProperties"] = "_target_" not in config and (
        "const" not in schema["properties"].get("_target_", {})
    )
    _config_dict = (
        OmegaConf.to_container(config, resolve=False)
        if isinstance(config, DictConfig)
        else config
    )
    assert isinstance(_config_dict, dict)
    for keys, value in _all_subentries_with_target(_config_dict).items():
        is_top_level: bool = not keys

        # logger.debug(f"Handling key {'.'.join(keys)} in config at path {config_file}")

        nested_value_schema = _get_schema_from_target(value)

        if "$defs" in nested_value_schema:
            # note: can't have a $defs key in the schema.
            schema.setdefault("$defs", {}).update(  # type: ignore
                nested_value_schema.pop("$defs")
            )
            assert "properties" in nested_value_schema

        if is_top_level:
            schema = merge_dicts(
                schema,
                nested_value_schema,
                conflict_handler=_overwrite,
                conflict_handlers=schema_conflict_handlers,
            )
            continue

        parent_keys, last_key = keys[:-1], keys[-1]
        where_to_set: Schema | ObjectSchema = schema
        for key in parent_keys:
            where_to_set = where_to_set.setdefault("properties", {}).setdefault(
                key, {"type": "object"}
            )  # type: ignore
            if "_target_" not in where_to_set:
                where_to_set["additionalProperties"] = True

        # logger.debug(f"Using schema from nested value at keys {keys}: {nested_value_schema}")

        if "properties" not in where_to_set:
            where_to_set["properties"] = {last_key: nested_value_schema}  # type: ignore
        elif last_key not in where_to_set["properties"]:
            assert isinstance(last_key, str)
            where_to_set["properties"][last_key] = nested_value_schema  # type: ignore
        else:
            where_to_set["properties"] = merge_dicts(  # type: ignore
                where_to_set["properties"],
                {last_key: nested_value_schema},  # type: ignore
                conflict_handler=_overwrite,
                conflict_handlers=schema_conflict_handlers,
            )

    return schema


def _update_schema_from_defaults(
    config_file: Path,
    schema: Schema,
    defaults: list[str | dict[str, str]],
    configs_dir: Path,
    repo_root: Path,
    config_store: ConfigStore | None,
):
    defaults_list = defaults
    config_store = config_store or ConfigStore.instance()

    for default in defaults_list:
        if default == "_self_":  # todo: does this actually make sense?
            continue
        # Note: The defaults can also have the .yaml or .yml extension, _load_config drops the
        # extension.
        default_config, default_config_file = load_default_config(
            config_file=config_file,
            configs_dir=configs_dir,
            default=default,
            repo_root=repo_root,
        )

        if "_target_" not in default_config:
            object_type = default_config._metadata.object_type
            logger.debug(f"{default=} {object_type=}")
            # TODO: if `object_type` is `dict`, this *may* be because the default config is:
            # {some_key: {some_value: thing_from_default}}
            if object_type is not dict:
                # todo: Figure out other examples where the object type is `dict` than the one above:
                # For now here we assume that the target is the same as the current one. Don't add one.
                # TODO: Need to check what the default is for object_type when `default_config` is not
                # a structured config.
                with omegaconf.open_dict(default_config):
                    default_config["_target_"] = (
                        object_type.__module__ + "." + object_type.__qualname__
                    )

        schema_of_default = _create_schema_for_config(
            config=default_config,
            # TODO: Something wrong about the `default_config_file` here. If the default
            # is something like {"a": {"b": some_config}} and default_config_file is "{config_dir/a/b/some_config}",
            # We shouldn't just recurse infinitely though!
            config_file=(
                default_config_file if isinstance(default, str) else config_file
            ),
            configs_dir=configs_dir,
            repo_root=repo_root,
            config_store=config_store,
        )
        # SUPER VERBOSE:
        # logger.debug(f"Schema from default {default}: {schema_of_default}")
        # logger.debug(
        #     f"Properties of {default=}: {list(schema_of_default['properties'].keys())}"
        # )  # type: ignore

        schema = merge_dicts(  # type: ignore
            schema_of_default,  # type: ignore
            schema,  # type: ignore
            conflict_handler=_overwrite,
            # conflict_handlers={
            #     "_target_": _overwrite,  # use the new target.
            #     "_target_.const": _overwrite,  # use the new target.
            #     "default": _overwrite,  # use the new default?
            #     "title": _overwrite,
            #     "description": _overwrite,
            # },
            conflict_handlers=schema_conflict_handlers,
        )
        # todo: deal with this one here.
        if schema.get("additionalProperties") is False:
            schema.pop("additionalProperties")
    return schema


def load_default_config(
    config_file: Path,
    configs_dir: Path,
    default: str | dict[str, str],
    repo_root: Path,
) -> tuple[DictConfig, Path]:
    """Loads the config pointed to by the given "defaults" entry in this config file.

    https://hydra.cc/docs/advanced/defaults_list/
    """
    assert "@" not in default  # TODO!

    if isinstance(default, str):
        assert not default.startswith("/")  # TODO!
        other_config_path = config_file.parent / default
    else:
        # TODO: It's more complicated that that! We need to also apply the default at the entry.
        # BUG: "override /db: something" should put the schema for `something` at key `db`!
        assert len(default) == 1
        key, val = next(iter(default.items()))
        key = key.removeprefix("override ")
        key = key.removeprefix("optional ")
        key = key.strip()
        assert key
        where_to_set = key.removeprefix("/").split("/")
        if key.startswith("/"):
            other_config_path = configs_dir / key.removeprefix("/") / val
        else:
            other_config_path = config_file.parent / key / val

        logger.debug(
            f"Loading config of default {default} with 'path': {pretty_path(other_config_path)}"
        )

        if not other_config_path.suffix:
            # If the other config file has the name without the extension, try both .yml and .yaml.
            for suffix in (".yml", ".yaml"):
                if other_config_path.with_suffix(suffix).exists():
                    other_config_path = other_config_path.with_suffix(suffix)
                    break
        default_config = load_config(
            other_config_path,
            configs_dir=configs_dir,
            repo_root=repo_root,
            config_store=ConfigStore.instance(),
        )
        assert where_to_set
        # if not where_to_set:
        #     return default_config, other_config_path

        default_config_result = {}
        result = default_config_result
        for i, parent in enumerate(where_to_set):
            if i == len(where_to_set) - 1:
                result[parent] = default_config
            else:
                result[parent] = {}
            result = result[parent]
        default_config = OmegaConf.create(default_config_result)
        logger.info(f"{default=}, {where_to_set=}, {default_config=}")
        # logger.info(f"{default_config=}, {where_to_set=}")
        return default_config, other_config_path

    logger.debug(
        f"Loading config of default {default} with 'path': {pretty_path(other_config_path)}"
    )

    if not other_config_path.suffix:
        # If the other config file has the name without the extension, try both .yml and .yaml.
        for suffix in (".yml", ".yaml"):
            if other_config_path.with_suffix(suffix).exists():
                other_config_path = other_config_path.with_suffix(suffix)
                break
    default_config = load_config(
        other_config_path,
        configs_dir=configs_dir,
        repo_root=repo_root,
        config_store=ConfigStore.instance(),
    )
    return (
        default_config,
        other_config_path,
    )


def _overwrite(val_a: Any, val_b: Any) -> Any:
    return val_b


def _keep_previous(val_a: Any, val_b: Any) -> Any:
    return val_a


def _has_package_global_line(config_file: Path) -> int | None:
    """Returns whether the config file contains a `@package _global_` directive of hydra.

    See: https://hydra.cc/docs/advanced/overriding_packages/#overriding-the-package-via-the-package-directive
    """
    for line in config_file.read_text().splitlines():
        line = line.strip()
        if not line.startswith("#"):
            continue
        if line.removeprefix("#").strip().startswith("@package _global_"):
            return True
    return False


# def _config_is_in_config_store(config_path: Path, config_store: ConfigStore) -> bool:
#     return config_store._open(str(config_path)) is not None


def _try_load_from_config_store(
    config_path: Path, configs_dir: Path, config_store: ConfigStore
) -> DictConfig | None:
    config_path = config_path.relative_to(configs_dir)

    def _config_is_in_config_store(config_path: Path) -> bool:
        return config_store._open(str(config_path)) is not None

    if _config_is_in_config_store(with_yml := config_path.with_suffix(".yml")):
        return config_store.load(str(with_yml)).node

    if _config_is_in_config_store(with_yaml := config_path.with_suffix(".yaml")):
        return config_store.load(str(with_yaml)).node

    return None


# NOTE: Tried to use `functools.cache` to prevent this from loading the plugins again, doesn't work
# import hydra._internal.utils
# hydra._internal.utils.create_config_search_path = functools.cache(
#     hydra._internal.utils.create_config_search_path
# )


# @functools.cache
def _create_config_search_path(search_path_dir: str | None) -> ConfigSearchPath:
    search_path = ConfigSearchPathImpl()
    search_path.append("hydra", "pkg://hydra.conf")

    if search_path_dir is not None:
        search_path.append("main", search_path_dir)

    search_path_plugins = Plugins.instance().discover(SearchPathPlugin)
    for spp in search_path_plugins:
        # CHANGED this, to avoid the weird re-instantiation of our plugin type.
        from hydra_plugins.auto_schema import auto_schema_plugin

        if spp is auto_schema_plugin.AutoSchemaPlugin:
            continue
        # TODO: This will still call all the other plugins once per call to this function!
        plugin = spp()
        assert isinstance(plugin, SearchPathPlugin)
        plugin.manipulate_search_path(search_path)

    search_path.append("schema", "structured://")

    return search_path


def load_config(
    config_path: Path,
    configs_dir: Path,
    repo_root: Path,
    config_store: ConfigStore,
) -> DictConfig:
    """Super overly-complicated function that tries to load a Hydra configuration file.

    This is in large part because Hydra's internal code is *very* complicated.
    """

    def _set_node_target(entry: ConfigNode | PossiblyNestedDict[str, ConfigNode]):
        if isinstance(entry, dict):
            for _key, value in entry.items():
                _set_node_target(value)
            return
        if "_target_" in entry.node:
            return

        target = entry.node._metadata.object_type
        if target is dict:
            assert entry.name == "_dummy_empty_config_.yaml"
            return

        logger.debug(
            f"Setting target for structured config node {entry.name} to {target}"
        )
        with omegaconf.open_dict(entry.node):
            target = f"{target.__module__}.{target.__qualname__}"
            entry.node["_target_"] = target

    for _key, entry in config_store.repo.items():
        _set_node_target(entry)

    if config := _try_load_from_config_store(
        config_path, configs_dir=configs_dir, config_store=config_store
    ):
        return config

    *config_groups, config_name = (
        config_path.relative_to(configs_dir).with_suffix("").parts
    )
    logger.debug(
        f"config_path: ./{pretty_path(config_path)}, {config_groups=}, {config_name=}, configs_dir: {configs_dir}"
    )
    config_group = "/".join(config_groups)

    from hydra.core.utils import setup_globals

    # todo: When would this normally be called?
    setup_globals()

    # FIXME!
    if configs_dir.is_relative_to(repo_root) and (configs_dir / "__init__.py").exists():
        config_module = str(configs_dir.relative_to(repo_root)).replace("/", ".")
        search_path = _create_config_search_path(f"pkg://{config_module}")
        logger.debug(f"Search path for a config module: {search_path}")
    else:
        search_path = _create_config_search_path(str(configs_dir))
        logger.debug(f"Search path for a config dir: {search_path}")

    config_loader = ConfigLoaderImpl(config_search_path=search_path)
    from hydra._internal.core_plugins.structured_config_source import (
        StructuredConfigSource,  # noqa
    )

    logger.debug(f"loading config {config_path}")

    # assert isinstance(
    #     schema_source := config_loader.repository.get_schema_source(),
    #     StructuredConfigSource,
    # )
    # assert False, schema_source.load_config(str(config_path))

    if _has_package_global_line(config_path):
        # Tricky: Here we load the global config but with the given config as an override.
        top_config = config_loader.load_configuration(
            "config",  # todo: Fix this!!
            overrides=[f"{config_group}={config_name}"],
            # todo: setting this here because it appears to be what's used in Hydra in a normal
            # run, even though RunMode.RUN would make more sense intuitively.
            run_mode=RunMode.MULTIRUN,
        )
        return top_config

    # Load the global config and get the node for the desired config.
    with warnings.catch_warnings():
        config_to_load = f"{config_group}/{config_name}"
        # if config_path.name == "config.yaml":
        #     # Make sure not to load the (base) Hydra config file that is also called `config.yaml`!
        #     config_to_load = str(config_path)
        warnings.simplefilter("ignore", category=UserWarning)
        top_config = config_loader.load_configuration(
            config_to_load,
            overrides=[],
            run_mode=RunMode.MULTIRUN,
            validate_sweep_overrides=False,
        )
    # Retrieve the sub-entry in the config and return it.
    config = top_config
    for config_group in config_groups:
        config = config[config_group]
    return config


def _add_schema_header(config_file: Path, schema_path: Path) -> None:
    """Add a comment in the yaml config file to tell yaml language server where to look for the
    schema.

    Importantly in the context of Hydra, this comment line should be added **after** any `#
    @package: <xyz>` directives of Hydra, otherwise Hydra doesn't use those directives properly
    anymore.
    """
    lines = config_file.read_text().splitlines(keepends=False)

    if config_file.parent is schema_path.parent:
        # TODO: Unsure when this branch would be used, and if it would differ.
        relative_path_to_schema = "./" + schema_path.name
    else:
        relative_path_to_schema = os.path.relpath(schema_path, start=config_file.parent)

    # Remove any existing schema lines.
    lines = [
        line
        for line in lines
        if not line.strip().startswith("# yaml-language-server: $schema=")
    ]

    # NOTE: This line can be placed anywhere in the file, not necessarily needs to be at the top,
    # and the yaml vscode extension will pick it up.
    new_line = f"# yaml-language-server: $schema={relative_path_to_schema}"

    package_global_line: int | None = None

    for i, line in enumerate(lines):
        line = line.strip()
        if not line:
            continue
        # BUG: IF the schema line comes before a @package: global comment, then the @package: _global_
        # comment is ignored by Hydra.
        # Locate the last package line (a bit unnecessary, since there should only be one).
        if line.startswith("#") and line.removeprefix("#").strip().startswith(
            "@package"
        ):
            package_global_line = i

    if package_global_line is None:
        # There's no package directive in the file.
        new_lines = [new_line, *lines]
    else:
        new_lines = lines.copy()
        # Insert the schema line after the package directive line.
        new_lines.insert(package_global_line + 1, new_line)

    result = "\n".join(new_lines).strip() + "\n"
    if config_file.read_text() != result:
        config_file.write_text(result)


def _get_dataclass_from_target(target: Any, config: dict | DictConfig) -> type:
    for target_type, special_kwargs in custom_hydra_zen_builds_args.items():
        if target_type is target or (
            inspect.isclass(target)
            and inspect.isclass(target_type)
            and issubclass(target, target_type)
        ):
            kwargs = merge_dicts(
                dict(
                    populate_full_signature=True,
                    hydra_recursive=False,
                    hydra_convert="all",
                    zen_dataclass={"cls_name": target.__qualname__},
                ),
                special_kwargs,
                conflict_handler=_overwrite,
            )
            # Generate the dataclass dynamically with hydra-zen.
            return hydra_zen.builds(target, **kwargs)
    if dataclasses.is_dataclass(target):
        # The target is a dataclass, so the schema is just the schema of the dataclass.
        assert inspect.isclass(target)
        return target
    # The target is a type or callable.
    assert callable(target)
    return hydra_zen.builds(
        target,
        populate_full_signature=True,
        hydra_defaults=config.get("defaults", None),
        hydra_recursive=False,
        hydra_convert="all",
        zen_dataclass={"cls_name": target.__qualname__.replace(".", "_")},
        # zen_wrappers=pydantic_parser,  # unsure if this is how it works?
    )


def _get_schema_from_target(config: dict | DictConfig) -> ObjectSchema | Schema:
    assert isinstance(config, dict | DictConfig)
    # logger.debug(f"Config: {config}")
    target = hydra.utils.get_object(config["_target_"])
    target_name = getattr(
        target, "__qualname__", getattr(target, "__name__", str(target))
    )
    object_type = _get_dataclass_from_target(target=target, config=config)

    try:
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)

            json_schema = pydantic.TypeAdapter(object_type).json_schema(
                mode="serialization",
                schema_generator=_MyGenerateJsonSchema,
                by_alias=False,
            )
            json_schema = typing.cast(Schema, json_schema)
        assert "properties" in json_schema
    except pydantic.PydanticSchemaGenerationError as e:
        raise NotImplementedError(f"Unable to get the schema with pydantic: {e}")

    assert "properties" in json_schema

    # Add a description
    json_schema["description"] = (
        f"Based on the signature of {target_name}.\n"
        + json_schema.get("description", "")
    )

    docs_to_search: list[dp.Docstring] = []

    if inspect.isclass(target):
        for target_or_base_class in inspect.getmro(target):
            if class_docstring := inspect.getdoc(target_or_base_class):
                docs_to_search.append(dp.parse(class_docstring))
            if init_docstring := inspect.getdoc(target_or_base_class.__init__):
                docs_to_search.append(dp.parse(init_docstring))
    else:
        assert inspect.isfunction(target) or inspect.ismethod(target), target
        docstring = inspect.getdoc(target)
        if docstring:
            docs_to_search = [dp.parse(docstring)]

    param_descriptions: dict[str, str] = {}
    for doc in docs_to_search:
        for param in doc.params:
            if param.description and param.arg_name not in param_descriptions:
                param_descriptions[param.arg_name] = param.description

    # Update the pydantic schema with descriptions:
    for property_name, property_dict in json_schema["properties"].items():
        if description := param_descriptions.get(property_name):
            property_dict["description"] = description
        else:
            property_dict[
                "description"
            ] = f"The {property_name} parameter of the {target_name}."

    if config.get("_partial_"):
        json_schema["required"] = []
    # Add some info on the target.
    if "_target_" not in json_schema["properties"]:
        json_schema["properties"]["_target_"] = {}
    else:
        assert isinstance(json_schema["properties"]["_target_"], dict)
    json_schema["properties"]["_target_"].update(
        PropertySchema(  # type: ignore
            type="string",
            title="Target",
            const=config["_target_"],
            # pattern=r"", # todo: Use a pattern to match python module import strings.
            description=(
                f"Target to instantiate, in this case: `{target.__name__}`\n"
                # f"* Source: <file://{relative_to_cwd(inspect.getfile(target))}>\n"
                # f"* Config file: <file://{config_file}>\n"
                f"See the Hydra docs for '_target_': https://hydra.cc/docs/advanced/instantiate_objects/overview/\n"
            ),
        )
    )

    # if the target takes **kwargs, then we don't restrict additional properties.
    json_schema["additionalProperties"] = (
        inspect.getfullargspec(target).varkw is not None
    )

    return json_schema


def _target_has_var_kwargs(config: DictConfig) -> bool:
    target = hydra_zen.get_target(config)  # type: ignore
    return inspect.getfullargspec(target).varkw is None


class _MyGenerateJsonSchema(GenerateJsonSchema):
    # def handle_invalid_for_json_schema(
    #     self, schema: core_schema.CoreSchema, error_info: str
    # ) -> JsonSchemaValue:
    #     raise PydanticOmit

    def enum_schema(self, schema: core_schema.EnumSchema) -> JsonSchemaValue:
        """Generates a JSON schema that matches an Enum value.

        Args:
            schema: The core schema.

        Returns:
            The generated JSON schema.
        """
        enum_type = schema["cls"]
        logger.debug(f"Getting the schema for Enum of type {enum_type}")
        for handler_enum_type, custom_handler in custom_enum_schemas.items():
            if handler_enum_type is enum_type or issubclass(
                enum_type, handler_enum_type
            ):
                schema = custom_handler(enum_type, schema)
                break
        return super().enum_schema(schema)



import dataclasses

from hydra.core.config_store import ConfigStore

cs = ConfigStore.instance()


@dataclasses.dataclass
class FooConfig:
    bar: str


# TODO: This still doesn't work!
cs.store(name="base_foo", node=FooConfig)



# import hydra_plugins.auto_schema.auto_schema_plugin  # noqa
# import hydra_plugins.auto_schema  # noqa
# import hydra_plugins  # noqa
import logging
from dataclasses import dataclass
from pathlib import Path

import hydra
import rich.logging
from hydra.core.config_store import ConfigStore
from omegaconf import MISSING, OmegaConf

from hydra_plugins.auto_schema import auto_schema_plugin

logging.basicConfig(
    level=logging.INFO,
    format=" %(message)s",
    datefmt="[%X]",
    force=True,
    handlers=[
        rich.logging.RichHandler(
            markup=True,
            rich_tracebacks=True,
            tracebacks_width=100,
            tracebacks_show_locals=False,
        )
    ],
)


@dataclass
class DBConfig:
    driver: str = MISSING
    host: str = "localhost"
    port: int = MISSING


@dataclass
class MySQLConfig(DBConfig):
    driver: str = "mysql"
    port: int = 3306
    user: str = MISSING
    password: str = MISSING


@dataclass
class PostGreSQLConfig(DBConfig):
    driver: str = "postgresql"
    user: str = MISSING
    port: int = 5432
    password: str = MISSING
    timeout: int = 10


@dataclass
class Config:
    db: DBConfig = MISSING
    debug: bool = False


if __name__ == "__main__":
    cs = ConfigStore.instance()
    cs.store(name="base_config", node=Config)
    cs.store(group="db", name="base_mysql", node=MySQLConfig)
    cs.store(group="db", name="base_postgresql", node=PostGreSQLConfig)


auto_schema_plugin.configure(
    auto_schema_plugin.AutoSchemaPluginConfig(
        schemas_dir=Path(__file__).parent / "schemas",
        regen_schemas=True,
        quiet=False,
        add_headers=True,
        verbose=True,
        stop_on_error=True,
    )
)


@hydra.main(version_base=None, config_path="conf", config_name="config")
def my_app(cfg: Config) -> None:
    print(cfg)
    print(OmegaConf.to_yaml(cfg, resolve=True))


if __name__ == "__main__":
    my_app()



import pytest
from hydra.core.singleton import Singleton


@pytest.fixture(autouse=True)
def reset_singletons_between_tests():
    state = Singleton.get_state()

    yield

    Singleton.set_state(state)



import datetime
import json
import logging
import warnings
from pathlib import Path

import rich
from hydra.core.config_store import ConfigStore
from watchdog.events import (
    DirCreatedEvent,
    DirDeletedEvent,
    DirModifiedEvent,
    DirMovedEvent,
    FileCreatedEvent,
    FileDeletedEvent,
    FileModifiedEvent,
    FileMovedEvent,
    PatternMatchingEventHandler,
)

from hydra_auto_schema.auto_schema import (
    _add_schema_header,
    _add_schemas_to_vscode_settings,
    _create_schema_for_config,
    _install_yaml_vscode_extension,
    _read_json,
    add_schemas_to_all_hydra_configs,
    get_schema_file_path,
    load_config,
    logger,
)

from .utils import pretty_path


class AutoSchemaEventHandler(PatternMatchingEventHandler):
    debounce_interval: datetime.timedelta = datetime.timedelta(seconds=3)

    def __init__(
        self,
        repo_root: Path,
        configs_dir: Path,
        schemas_dir: Path,
        regen_schemas: bool,
        stop_on_error: bool,
        quiet: bool,
        add_headers: bool | None,
        config_store: ConfigStore | None = None,
    ):
        self.configs_dir = configs_dir
        super().__init__(
            patterns=[
                f"{self.configs_dir}/**.yaml",
                f"{self.configs_dir}/**.yml",
                f"{self.configs_dir}/**/*.yaml",
                f"{self.configs_dir}/**/*.yml",
            ],
            ignore_patterns=None,
            ignore_directories=True,
            case_sensitive=True,
        )
        self.repo_root = repo_root
        self.schemas_dir = schemas_dir
        self.regen_schemas = regen_schemas
        self.stop_on_error = stop_on_error
        self.quiet = quiet
        self.add_headers = add_headers

        self._last_updated: dict[Path, datetime.datetime] = {}
        # On startup, we could make a schema for every config file, right?
        add_schemas_to_all_hydra_configs(
            repo_root=repo_root,
            configs_dir=configs_dir,
            schemas_dir=schemas_dir,
            regen_schemas=regen_schemas,
            stop_on_error=stop_on_error,
            quiet=quiet,
            add_headers=add_headers,
            config_store=config_store or ConfigStore.instance(),
        )
        self.console = rich.console.Console()
        self.console.log(
            f"Watching for changes in config files in the {pretty_path(configs_dir)} directory."
        )

    def on_created(self, event: DirCreatedEvent | FileCreatedEvent) -> None:
        logger.debug(f"on_created event: {event.src_path}")
        if config_file := self._filter_config_file(event.src_path):
            logger.info(f"Config file was created: {config_file}")
            self.run(config_file)

    def on_deleted(self, event: DirDeletedEvent | FileDeletedEvent) -> None:
        logger.debug(f"on_deleted event: {event.src_path}")

        # TODO: Remove the schema from the vscode settings.
        if config_file := self._filter_config_file(event.src_path):
            logger.info(f"Config file was deleted: {config_file}")
            self._remove_schema_file(config_file)

    def on_modified(self, event: DirModifiedEvent | FileModifiedEvent) -> None:
        logger.debug(f"on_modified event: {event.src_path}")

        if config_file := self._filter_config_file(event.src_path):
            if self._debounce(config_file):
                return
            logger.info(f"Config file was modified: {config_file}")
            self.run(config_file)

    def on_moved(self, event: DirMovedEvent | FileMovedEvent) -> None:
        logger.debug(f"on_moved event: {event.src_path}")

        source_file = self._filter_config_file(event.src_path)
        dest_file = self._filter_config_file(event.dest_path)
        if source_file and dest_file:
            logger.info(f"Config file was moved from {source_file} --> {dest_file}")
            self._remove_schema_file(source_file)
            self.run(dest_file)

        if dest_file:
            self.run(dest_file)

    def _filter_config_file(self, p: str | bytes) -> Path | None:
        config_file = Path(str(p))
        if config_file.suffix not in [".yaml", ".yml"]:
            return None
        if not config_file.is_relative_to(self.configs_dir):
            return None
        return config_file

    def _debounce(self, config_file: Path) -> bool:
        """Debouncing to prevent a loop caused by us modifying the config file to add a header(?).

        TODO: Also seems to be necessary when `self.add_header` is False! Why?

        Returns:
            `True` if the file was modified recently (by us), `False` otherwise.
        """
        if config_file not in self._last_updated:
            self._last_updated[config_file] = datetime.datetime.now()
            return False

        if (
            datetime.datetime.now() - self._last_updated[config_file]
        ) < self.debounce_interval:
            logger.debug(
                f"Config file {config_file} was modified very recently; not regenerating the schema."
            )
            return True
        self._last_updated[config_file] = datetime.datetime.now()
        return False

    def run(self, config_file: Path) -> None:
        p = pretty_path(config_file)
        try:
            self._run(config_file)
        except Exception as exc:
            logger.warning(f"Error while processing config at {config_file}:\n{exc}")
            if self.stop_on_error:
                raise
            sees_warning = logger.getEffectiveLevel() <= logging.WARNING
            self.console.log(
                f"Unable to generate the schema for {p}."
                + ("" if sees_warning else " (use -v for more info).")
            )
        else:
            self.console.log(f"Schema updated for {p}.")

    def _run(self, config_file: Path) -> None:
        pretty_config_file_name = config_file.relative_to(self.configs_dir)
        schema_file = get_schema_file_path(config_file, self.schemas_dir)

        logger.debug(f"Creating a schema for {pretty_config_file_name}")
        from hydra.core.config_store import ConfigStore

        config_store = ConfigStore.instance()

        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            config = load_config(
                config_file,
                configs_dir=self.configs_dir,
                repo_root=self.repo_root,
                config_store=config_store,
            )
        schema = _create_schema_for_config(
            config,
            config_file=config_file,
            configs_dir=self.configs_dir,
            repo_root=self.repo_root,
            config_store=config_store,
        )
        schema_file.parent.mkdir(exist_ok=True, parents=True)
        schema_file.write_text(json.dumps(schema, indent=2).rstrip() + "\n\n")
        # _set_is_incomplete_schema(schema_file, False)

        if not self.add_headers:
            try:
                _install_yaml_vscode_extension()
            except OSError:
                pass

            try:
                _add_schemas_to_vscode_settings(
                    {config_file: schema_file}, repo_root=self.repo_root
                )
            except Exception as exc:
                logger.error(
                    f"Unable to write schemas in the vscode settings file. "
                    f"Falling back to adding a header to config files. (exc={exc})"
                )
                if self.add_headers is not None:
                    # Unable to do it. Don't try to add headers, just return.
                    return
            else:
                # Success. Return.
                return

        logger.debug(
            "Adding a header to the config file to point to the schema to use."
        )
        _add_schema_header(config_file, schema_path=schema_file)

    def _remove_schema_file(self, config_file: Path) -> None:
        schema_file = get_schema_file_path(config_file, self.schemas_dir)
        if self.add_headers:
            # Could also remove the schema file for this config file.
            logger.debug(
                f"Removing schema file associated with config {config_file}: {schema_file}"
            )
            schema_file.unlink()
        else:
            vscode_settings_file = self.repo_root / ".vscode" / "settings.json"
            if not vscode_settings_file.exists():
                # Weird.
                return
            vscode_settings = _read_json(vscode_settings_file)
            yaml_schemas = vscode_settings.get("yaml.schemas", {})
            assert isinstance(yaml_schemas, dict)
            if not yaml_schemas:
                # Weird also!
                return
            if (
                schema_key := str(schema_file.relative_to(self.repo_root))
            ) not in yaml_schemas:
                # Weird also!
                return
            config_file_value = str(config_file)
            configs_with_schema: str | list[str] = yaml_schemas[schema_key]
            if isinstance(configs_with_schema, str):
                assert configs_with_schema == config_file_value
                # Weird!
                # Remove the schema from the dict.
                yaml_schemas.pop(schema_key)
            else:
                yaml_schemas[schema_key].remove(config_file_value)
            # NOTE: This doesn't work with comments.
            vscode_settings_file.write_text(json.dumps(vscode_settings, indent=4))



"""Global variables that can be used to customize how schemas are generated."""
import dataclasses
import enum
from typing import Any, Callable

from pydantic.json_schema import JsonSchemaValue
from pydantic_core import core_schema

custom_hydra_zen_builds_args: dict[type | Callable, dict] = {
    # flax.linen.Module: {"zen_exclude": ["parent"]},
    # lightning.pytorch.callbacks.RichProgressBar: {"zen_exclude": ["theme"]},
}
"""Keyword arguments that should be passed to `hydra_zen.builds` for a given class or callable.

These arguments overwrite the default values.
"""

custom_enum_schemas: dict[type[enum.Enum], Callable] = {}
"""Dict of functions to be used by pydantic to generate schemas for enum classes.

TODO: This a bit too specific. We could probably use our `GenerateJsonSchema`
subclass to enable more general customizations, following this guide:
https://docs.pydantic.dev/2.9/concepts/json_schema/#customizing-the-json-schema-generation-process
"""

schema_conflict_handlers: dict[str, Callable[[Any, Any], Any]] = {}
"""Functions to be used by the `merge_dicts` function to resolve conflicts between schemas.

See the docstring of `merge_dicts` for more info.
"""

# Conditionally add some common fixes?
# TODO: Should we really import those here? This might make the import pretty slow, no? If not,
# when / how should we import things here?

try:
    from flax.linen import Module  # type: ignore

    custom_hydra_zen_builds_args[Module] = {"zen_exclude": ["parent"]}
except ImportError:
    pass

try:
    from lightning.pytorch.callbacks import RichProgressBar  # type: ignore

    custom_hydra_zen_builds_args[RichProgressBar] = {"zen_exclude": ["theme"]}
except ImportError:
    pass


try:
    from torchvision.models import WeightsEnum as _WeightsEnum  # type: ignore

    def _handle_torchvision_weights_enum(
        enum_type: type[_WeightsEnum], schema: core_schema.EnumSchema
    ) -> JsonSchemaValue:
        @dataclasses.dataclass
        class Dummy:
            value: str

        slightly_changed_schema = schema | {
            "members": [Dummy(v.name) for v in schema["members"]]
        }
        return slightly_changed_schema

    custom_enum_schemas[_WeightsEnum] = _handle_torchvision_weights_enum
except ImportError:
    pass



import importlib
import logging
import typing
from pathlib import Path
from typing import ClassVar

import hydra.core.plugins
import hydra_zen
from hydra.core.config_search_path import ConfigSearchPath
from hydra.core.config_store import ConfigStore  # noqa
from hydra.plugins.search_path_plugin import SearchPathPlugin

from hydra_auto_schema.auto_schema import add_schemas_to_all_hydra_configs, logger

# TODO: Perhaps this should be a different kind of plugin, one that has access to the structured
# schemas within the hydra app, so it could be executed later when calling the `hydra.main`
# function?


def register_auto_schema_plugin():
    hydra.core.plugins.Plugins.instance().register(AutoSchemaPlugin)


@hydra_zen.hydrated_dataclass(
    add_schemas_to_all_hydra_configs,
    populate_full_signature=True,
    zen_partial=True,
    zen_meta={"verbose": False},
)
class AutoSchemaPluginConfig:
    """Config for the AutoSchemaPlugin."""

    schemas_dir: Path | None = None
    regen_schemas: bool = False
    stop_on_error: bool = False
    quiet: bool = True
    add_headers: bool | None = False
    if typing.TYPE_CHECKING:
        verbose: bool = False


config: AutoSchemaPluginConfig | None
"""Global config used by default by the AutoSchemaPlugin."""

# is called by hydra with the plugin module (even if we've already imported it!)
# HACK: Prevent a second exec of this module from changing this value.
# This happens in the `hydra.core.plugins.Plugins._scan_all_plugins` method, where `exec_module`
if "config" not in globals():
    config = None


def configure(new_config: AutoSchemaPluginConfig) -> None:
    global config
    logger.debug(f"Setting the configuration for the AutoSchemaPlugin: {new_config}")
    config = new_config


class AutoSchemaPlugin(SearchPathPlugin):
    provider: str
    path: str

    _ALREADY_DID: ClassVar[bool] = False

    def __init__(self) -> None:
        super().__init__()
        global config
        self.config = config if config is not None else AutoSchemaPluginConfig()
        self.fn = hydra_zen.instantiate(self.config)
        self.cs = ConfigStore.instance()

        logger.debug(
            f"The AutoSchemaPlugin is being instantiated with the following config: {self.config}"
        )
        if self.config.verbose:
            from hydra_auto_schema.auto_schema import logger as _logger

            _logger.setLevel(logging.INFO)

    def manipulate_search_path(self, search_path: ConfigSearchPath) -> None:
        # IDEA: Try to get the configs dir and the repo root from the search path, and call the
        # function on the config files in that directory.
        if type(self)._ALREADY_DID:
            # TODO: figure out what's causing this.
            logger.debug(f"Avoiding weird multiple calls to the {AutoSchemaPlugin}!")
            return
        type(self)._ALREADY_DID = True

        search_path_entries = search_path.get_path()

        # WIP: Trying to infer the project root, configs dir from the Hydra context.
        # Currently hard-coded.

        for search_path_entry in search_path_entries:
            # TODO: There are probably lots of assumptions that are specific to the
            # ResearchTemplate repo that would need to be removed / generalized before we can make
            # this a pip-installable hydra plugin..
            logger.debug(f"search_path_entry: {search_path_entry}")
            if search_path_entry.provider != "main":
                continue

            if search_path_entry.path.startswith("pkg://"):
                configs_pkg = search_path_entry.path.removeprefix("pkg://")
                project_package = configs_pkg.split(".")[0].replace("/", ".")
                logger.debug(f"Found project package: {project_package}")
                _project_module = importlib.import_module(project_package)
                # FIXME: THIS IS SPECIFIC TO THE RESEARCH TEMPLATE!
                assert (
                    _project_module.__file__
                    and Path(_project_module.__file__).name == "__init__.py"
                )
                repo_root = Path(_project_module.__file__).parent.parent
                configs_dir = repo_root / configs_pkg.replace(".", "/")
            else:
                # TODO: Check for file:// prefixes?
                configs_dir = Path(search_path_entry.path)
                assert configs_dir.is_dir(), configs_dir
                repo_root = configs_dir.parent  # wild assumption, no?

            if not (repo_root.is_dir() and configs_dir.is_dir()):
                logger.warning(
                    f"Unable to add schemas to Hydra configs: "
                    f"Expected to find the project root directory at {repo_root} "
                    f"and the configs directory at {configs_dir}!"
                )
                continue
            self.fn(
                repo_root=repo_root,
                configs_dir=configs_dir,
                config_store=self.cs,
            )



""" TODO: Tests for getting the schema from structured configs. """

import shlex
import shutil
import subprocess
from pathlib import Path

import pytest
from pytest_regressions.file_regression import FileRegressionFixture

from hydra_plugins.auto_schema import auto_schema_plugin

structured_app_dir = Path(__file__).parent / "structured_app"


@pytest.fixture
def new_repo_root(tmp_path: Path):
    new_repo_root = tmp_path / structured_app_dir.name
    shutil.copytree(structured_app_dir, new_repo_root)

    return new_repo_root


@pytest.fixture(params=[True, False])
def schemas_already_exist(request: pytest.FixtureRequest):
    return request.param


@pytest.fixture
def new_schemas_dir(new_repo_root: Path, schemas_already_exist: bool):
    schemas_dir = new_repo_root / "schemas"
    if not schemas_already_exist:
        shutil.rmtree(schemas_dir, ignore_errors=True)
    return schemas_dir


@pytest.fixture()
def set_config(
    tmp_path: Path,
    new_schemas_dir: Path,
    new_repo_root: Path,
    monkeypatch: pytest.MonkeyPatch,
):
    monkeypatch.setattr(
        auto_schema_plugin,
        "config",
        auto_schema_plugin.AutoSchemaPluginConfig(
            schemas_dir=new_schemas_dir,
            add_headers=True,
            regen_schemas=True,
            stop_on_error=True,
            quiet=False,
            verbose=True,
        ),
    )


@pytest.fixture
def command_line_arguments(request: pytest.FixtureRequest):
    return getattr(request, "param", "")


@pytest.fixture
def structured_app_result(
    new_repo_root: Path,
    set_config: None,
    command_line_arguments: str,
    monkeypatch: pytest.MonkeyPatch,
):
    # monkeypatch.chdir(new_repo_root)
    result = subprocess.run(
        ["python", str(new_repo_root / "app.py"), *shlex.split(command_line_arguments)],
        capture_output=True,
        text=True,
    )
    assert result.returncode == 1
    print(result.stdout)
    print(result.stderr)
    return result


@pytest.mark.parametrize(
    command_line_arguments.__name__, ["db.port=fail"], indirect=True
)
def test_run_example(
    structured_app_result: subprocess.CompletedProcess,
    new_repo_root: Path,
    new_schemas_dir: Path,
    file_regression: FileRegressionFixture,
):
    assert structured_app_result.returncode == 1
    assert (
        "Value 'fail' of type 'str' could not be converted to Integer"
        in structured_app_result.stderr
    )
    assert "full_key: db.port" in structured_app_result.stderr
    assert "reference_type=DBConfig" in structured_app_result.stderr
    assert "object_type=MySQLConfig" in structured_app_result.stderr
    # The schemas should have been generated.
    schemas_dir = new_schemas_dir
    assert schemas_dir.exists()
    files = list(schemas_dir.glob("*.json"))
    assert files
    for file in files:
        file_regression.check(
            (schemas_dir / file.name).read_text().rstrip(),
            extension=".json",
            fullpath=structured_app_dir / "schemas" / file.name,
        )


@pytest.mark.xfail(reason="TODO: implement the test.", strict=True)
@pytest.mark.parametrize(
    command_line_arguments.__name__, ["--config-name=with_overrides"], indirect=True
)
def test_run_example_with_overrides(
    structured_app_result: subprocess.CompletedProcess,
    new_schemas_dir: Path,
    file_regression: FileRegressionFixture,
):
    assert structured_app_result.returncode == 1
    assert False, structured_app_result.stdout
    # The schemas should have been generated.
    schemas_dir = new_schemas_dir
    assert schemas_dir.exists()
    files = list(schemas_dir.glob("*.json"))
    assert files
    for file in files:
        file_regression.check(
            (schemas_dir / file.name).read_text().rstrip(),
            extension=".json",
            fullpath=structured_app_dir / "schemas" / file.name,
        )



# from ..hydra_plugins.auto_schema.auto_schema_plugin import (
#     AutoSchemaPlugin,
#     register_auto_schema_plugin,
# )
from .auto_schema import (
    add_schemas_to_all_hydra_configs,
)
from .customize import custom_enum_schemas, custom_hydra_zen_builds_args
from .filewatcher import AutoSchemaEventHandler

__all__ = [
    "add_schemas_to_all_hydra_configs",
    # "AutoSchemaPlugin",
    # "register_auto_schema_plugin",
    "AutoSchemaEventHandler",
    "custom_hydra_zen_builds_args",
    "custom_enum_schemas",
]



import shlex
import subprocess
import warnings
from pathlib import Path

import hydra.core.plugins
import hydra.errors
import pytest
from hydra.plugins.search_path_plugin import SearchPathPlugin

from hydra_auto_schema.__main__ import main
from hydra_plugins.auto_schema.auto_schema_plugin import (
    AutoSchemaPlugin,
)

this_dir = Path(__file__).parent
test_configs_dir = this_dir / "configs"
structured_app_dir = this_dir / "structured_app"


@pytest.mark.xfail(
    raises=hydra.errors.MissingConfigException,
    reason="Doesn't work yet, need to get a populated Hydra ConfigStore outside the Hydra context.",
    strict=True,
)
@pytest.mark.parametrize(
    "args",
    [
        # f"{structured_app_dir} --stop-on-error --regen-schemas -vvv",
        f"{this_dir} --stop-on-error --regen-schemas -vvv",
        # TODO: Perhaps we could try to import the repo_root as a python module, or add it to
        # path, or something similar, so that we can actually get a "hydrated" ConfigStore object?
        f"{structured_app_dir} --stop-on-error --regen-schemas -vvv",
    ],
)
def test_run_via_cli_without_errors(args: str):
    """Checks that the command completes without errors."""
    # Run programmatically instead of with a subprocess so we can get nice coverage stats.
    # assuming we're at the project root directory.
    from hydra.core.config_store import ConfigStore

    warnings.warn(RuntimeWarning(ConfigStore.instance().repo.keys()))
    main(shlex.split(args))


@pytest.mark.skip(reason="Shouldn't really be run as a uv tool anyway.")
def test_run_with_uvx():
    """Actually run the command on the repo, via the `[tool.rye.scripts]` entry in
    pyproject.toml."""
    # Run once so we can get nice coverage stats.
    subprocess.check_call(
        [
            "uvx",
            "--from=.",
            "--reinstall-package=hydra-auto-schema",
            "hydra-auto-schema",
            f"{test_configs_dir}",
        ],
        text=True,
    )


@pytest.mark.skip(reason="Shouldn't really be run as a uv tool anyway.")
def test_run_as_uv_tool():
    """Actually run the command on the repo, via the `[tool.rye.scripts]` entry in
    pyproject.toml."""
    # Run once so we can get nice coverage stats.
    subprocess.check_call(
        [
            "uv",
            "tool",
            "run",
            "--from=.",
            "--reinstall-package=hydra-auto-schema",
            "hydra-auto-schema",
            f"{test_configs_dir}",
        ],
        text=True,
    )


# @pytest.mark.xfail(
#     reason="Turning off the plugin discovery for now while it's not ready."
# )
def test_plugin_is_discoverable(monkeypatch: pytest.MonkeyPatch):
    # NOTE: This is super bad, it executes the plugin modules, and inserts them forcefully into
    # `sys.modules`, which overwrites our loaded modules!
    # class PreventOverwriteByHydraPlugins(dict):
    #     def __setitem__(self, key: str, value) -> None:
    #         if key.startswith("hydra_plugins.") and key in self:
    #             return
    #         super().__setitem__(key, value)

    # can't seem to be able to import `Loader` of importlib. :(
    # monkeypatch.setattr(importlib.Loader, "exec_module", lambda module: None)
    # monkeypatch.setattr(sys, "modules", PreventOverwriteByHydraPlugins(sys.modules))
    plugins = hydra.core.plugins.Plugins.instance().discover(SearchPathPlugin)
    # assert AutoSchemaPlugin in plugins
    assert AutoSchemaPlugin.__name__ in [p.__name__ for p in plugins]



from __future__ import annotations
from collections.abc import Mapping, MutableMapping, Sequence
from typing import Any, Literal, TypedDict
from typing_extensions import NotRequired, Required


_SpecialEntries = TypedDict(
    "_SpecialEntries", {"$defs": dict[str, dict], "$schema": str}, total=False
)


class PropertySchema(_SpecialEntries, total=False):
    title: str
    type: str
    description: str
    default: Any
    examples: list[str]
    deprecated: bool
    readOnly: bool
    writeOnly: bool
    const: Any
    enum: list[Any]


class OneOf(TypedDict):
    oneOf: Sequence[
        PropertySchema | ArrayPropertySchema | ObjectSchema | StringPropertySchema
    ]


class ArrayPropertySchema(PropertySchema, total=False):
    type: Literal["array"]
    items: Required[PropertySchema | OneOf]
    minItems: int
    maxItems: int
    uniqueItems: bool


class StringPropertySchema(PropertySchema, total=False):
    type: Literal["string"]
    pattern: str


class _PropertyNames(TypedDict):
    pattern: str


class ObjectSchema(PropertySchema, total=False):
    type: Literal["object"]
    # annoying that we have to include the subclasses here!
    properties: MutableMapping[
        str, PropertySchema | ArrayPropertySchema | StringPropertySchema | ObjectSchema
    ]
    patternProperties: Mapping[str, PropertySchema | StringPropertySchema]
    propertyNames: _PropertyNames
    minProperties: int
    maxProperties: int


class Schema(TypedDict, total=False):
    # "$defs":
    title: str
    description: str
    type: str

    # annoying that we have to include the subclasses here!
    properties: Required[
        MutableMapping[
            str,
            PropertySchema | ArrayPropertySchema | StringPropertySchema | ObjectSchema,
        ]
    ]
    required: NotRequired[list[str]]

    additionalProperties: NotRequired[bool]

    dependentRequired: NotRequired[MutableMapping[str, list[str]]]
    """ https://json-schema.org/understanding-json-schema/reference/conditionals#dependentRequired """


HYDRA_CONFIG_SCHEMA = Schema(
    title="Default Schema for any Hydra config file.",
    description="Schema created by the `auto_schema.py` script.",
    properties={
        "defaults": ArrayPropertySchema(
            title="Hydra defaults",
            description="Hydra defaults for this config. See https://hydra.cc/docs/advanced/defaults_list/",
            type="array",
            items=OneOf(
                oneOf=[
                    ObjectSchema(
                        type="object",
                        propertyNames={"pattern": r"^(override\s*)?(/?\w*)+$"},
                        patternProperties={
                            # todo: support package directives with @?
                            # todo: Create a enum schema using the available options in the config group!
                            # override network: something  -> `something` value should be in the available options for network.
                            r"^(override\s*)?(/?\w*)*$": StringPropertySchema(
                                type="string", pattern=r"\w*(.yaml|.yml)?$"
                            ),
                        },
                        minProperties=1,
                        maxProperties=1,
                    ),
                    StringPropertySchema(type="string", pattern=r"^\w+(.yaml|.yml)?$"),
                    ObjectSchema(
                        type="object",
                        propertyNames=_PropertyNames(
                            pattern=r"^(override\s*)?(/?\w*)+$"
                        ),
                        patternProperties={
                            r"^(override\s*)?(/?\w*)*$": PropertySchema(type="null"),
                        },
                        minProperties=1,
                        maxProperties=1,
                    ),
                ],
            ),
            uniqueItems=True,
        ),
        "_target_": StringPropertySchema(
            type="string",
            title="Target",
            description="Target to instantiate.\nSee https://hydra.cc/docs/advanced/instantiate_objects/overview/",
        ),
        "_convert_": StringPropertySchema(
            type="string",
            enum=["none", "partial", "object", "all"],
            title="Convert",
            description="See https://hydra.cc/docs/advanced/instantiate_objects/overview/#parameter-conversion-strategies",
        ),
        "_partial_": PropertySchema(
            type="boolean",
            title="Partial",
            description=(
                "Whether this config calls the target function when instantiated, or creates "
                "a `functools.partial` that will call the target.\n"
                "See: https://hydra.cc/docs/advanced/instantiate_objects/overview"
            ),
        ),
        "_recursive_": PropertySchema(
            type="boolean",
            title="Recursive",
            description=(
                "Whether instantiating this config should recursively instantiate children configs.\n"
                "See: https://hydra.cc/docs/advanced/instantiate_objects/overview/#recursive-instantiation"
            ),
        ),
    },
    dependentRequired={
        "_convert_": ["_target_"],
        "_partial_": ["_target_"],
        "_args_": ["_target_"],
        "_recursive_": ["_target_"],
    },
)



import enum
import functools
import json
import unittest
import unittest.mock
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable
from unittest.mock import Mock

import hydra_zen
import omegaconf
import pydantic_core
import pytest
from pytest_regressions.file_regression import FileRegressionFixture

from hydra_auto_schema.auto_schema import _create_schema_for_config
from hydra_auto_schema.utils import get_target_string

## Copied from torchvision.


@dataclass
class Weights:
    """This class is used to group important attributes associated with the pre-trained weights.

    Args:
        url (str): The location where we find the weights.
        transforms (Callable): A callable that constructs the preprocessing method (or validation preset transforms)
            needed to use the model. The reason we attach a constructor method rather than an already constructed
            object is because the specific object might have memory and thus we want to delay initialization until
            needed.
        meta (Dict[str, Any]): Stores meta-data related to the weights of the model and its configuration. These can be
            informative attributes (for example the number of parameters/flops, recipe link/methods used in training
            etc), configuration parameters (for example the `num_classes`) needed to construct the model or important
            meta-data (for example the `classes` of a classification model) needed to use the model.
    """

    url: str
    transforms: Callable
    meta: dict[str, Any]

    def __eq__(self, other: Any) -> bool:
        # We need this custom implementation for correct deep-copy and deserialization behavior.
        # TL;DR: After the definition of an enum, creating a new instance, i.e. by deep-copying or deserializing it,
        # involves an equality check against the defined members. Unfortunately, the `transforms` attribute is often
        # defined with `functools.partial` and `fn = partial(...); assert deepcopy(fn) != fn`. Without custom handling
        # for it, the check against the defined members would fail and effectively prevent the weights from being
        # deep-copied or deserialized.
        # See https://github.com/pytorch/vision/pull/7107 for details.
        if not isinstance(other, Weights):
            return NotImplemented

        if self.url != other.url:
            return False

        if self.meta != other.meta:
            return False

        if isinstance(self.transforms, functools.partial) and isinstance(
            other.transforms, functools.partial
        ):
            return (
                self.transforms.func == other.transforms.func
                and self.transforms.args == other.transforms.args
                and self.transforms.keywords == other.transforms.keywords
            )
        else:
            return self.transforms == other.transforms


# C
class WeightsEnum(enum.Enum):
    """This class is the parent class of all model weights. Each model building method receives an
    optional `weights` parameter with its associated pre-trained weights. It inherits from `Enum`
    and its values should be of type `Weights`.

    Args:
        value (Weights): The data class entry with the weight information.
    """

    @classmethod
    def verify(cls, obj: Any) -> Any:
        if obj is not None:
            if type(obj) is str:
                obj = cls[obj.replace(cls.__name__ + ".", "")]
            elif not isinstance(obj, cls):
                raise TypeError(
                    f"Invalid Weight class provided; expected {cls.__name__} but received {obj.__class__.__name__}."
                )
        return obj

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}.{self._name_}"

    @property
    def url(self):
        return self.value.url

    @property
    def transforms(self):
        return self.value.transforms

    @property
    def meta(self):
        return self.value.meta


class InterpolationMode(enum.Enum):
    BILINEAR = enum.auto()


class ImageClassification:
    def __init__(
        self,
        *,
        crop_size: int,
        resize_size: int = 256,
        mean: tuple[float, ...] = (0.485, 0.456, 0.406),
        std: tuple[float, ...] = (0.229, 0.224, 0.225),
        interpolation: InterpolationMode = InterpolationMode.BILINEAR,
        antialias: bool | None = True,
    ) -> None:
        super().__init__()


_COMMON_META = {
    "min_size": (1, 1),
    "categories": [...],
}


class ResNet50_Weights(WeightsEnum):
    IMAGENET1K_V1 = Weights(
        url="https://download.pytorch.org/models/resnet50-0676ba61.pth",
        transforms=functools.partial(ImageClassification, crop_size=224),
        meta={
            **_COMMON_META,
            "num_params": 25557032,
            "recipe": "https://github.com/pytorch/vision/tree/main/references/classification#resnet",
            "_metrics": {
                "ImageNet-1K": {
                    "acc@1": 76.130,
                    "acc@5": 92.862,
                }
            },
            "_ops": 4.089,
            "_file_size": 97.781,
            "_docs": """These weights reproduce closely the results of the paper using a simple training recipe.""",
        },
    )
    IMAGENET1K_V2 = Weights(
        url="https://download.pytorch.org/models/resnet50-11ad3fa6.pth",
        transforms=functools.partial(
            ImageClassification, crop_size=224, resize_size=232
        ),
        meta={
            **_COMMON_META,
            "num_params": 25557032,
            "recipe": "https://github.com/pytorch/vision/issues/3995#issuecomment-1013906621",
            "_metrics": {
                "ImageNet-1K": {
                    "acc@1": 80.858,
                    "acc@5": 95.434,
                }
            },
            "_ops": 4.089,
            "_file_size": 97.79,
            "_docs": """
                These weights improve upon the results of the original paper by using TorchVision's `new training recipe
                <https://pytorch.org/blog/how-to-train-state-of-the-art-models-using-torchvision-latest-primitives/>`_.
            """,
        },
    )
    DEFAULT = IMAGENET1K_V2


def resnet50(
    *, weights: ResNet50_Weights | None = None, progress: bool = True, **kwargs: Any
) -> Any:
    return {"weights": weights, "progress": progress, **kwargs}


def test_use_custom_enum_handler(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    file_regression: FileRegressionFixture,
):
    from hydra_auto_schema.customize import custom_enum_schemas

    def _handle_torchvision_weights_enum(enum_type: type[WeightsEnum], schema):
        @dataclass
        class Dummy:
            value: str

        slightly_changed_schema = schema | {
            "members": [Dummy(v.name) for v in schema["members"]]
        }
        return slightly_changed_schema

    mock = Mock(
        spec=_handle_torchvision_weights_enum, wraps=_handle_torchvision_weights_enum
    )

    config = {
        "_target_": get_target_string(resnet50),
    }
    config_file = tmp_path / "config.yaml"
    omegaconf.OmegaConf.save(config, config_file)
    with pytest.raises(
        pydantic_core._pydantic_core.PydanticSerializationError,
        match="Unable to serialize unknown type: <class 'functools.partial'>",
    ):
        _ = _create_schema_for_config(
            config,
            config_file,
            configs_dir=tmp_path,
            repo_root=tmp_path,
            config_store=None,
        )
    mock.assert_not_called()

    mock.reset_mock()

    monkeypatch.setitem(custom_enum_schemas, WeightsEnum, mock)

    schema = _create_schema_for_config(
        config,
        config_file,
        configs_dir=tmp_path,
        repo_root=tmp_path,
        config_store=None,
    )
    mock.assert_called_once_with(ResNet50_Weights, unittest.mock.ANY)

    file_regression.check(
        json.dumps(schema, indent=2) + "\n",
        extension=".json",
    )


class A:
    def __init__(self, a: int = 1):
        self.a = a


def test_use_custom_builds_kwargs(
    monkeypatch: pytest.MonkeyPatch,
    tmp_path: Path,
    file_regression: FileRegressionFixture,
):
    from hydra_auto_schema.customize import custom_hydra_zen_builds_args

    mock_hydra_zen_builds = Mock(spec=hydra_zen.builds, wraps=hydra_zen.builds)
    monkeypatch.setattr(hydra_zen, hydra_zen.builds.__name__, mock_hydra_zen_builds)

    config = {"_target_": get_target_string(A)}
    config_file = tmp_path / "config.yaml"
    omegaconf.OmegaConf.save(config, config_file)

    _schema = _create_schema_for_config(
        config,
        config_file,
        configs_dir=tmp_path,
        repo_root=tmp_path,
        config_store=None,
    )

    mock_hydra_zen_builds.assert_called_once()
    call = mock_hydra_zen_builds.call_args_list[0]
    assert call.args[0] is A
    # The `zen_exclude` kwarg shouldn't be in the call args to start with.
    assert "zen_exclude" not in call.kwargs

    mock_hydra_zen_builds.reset_mock()

    # Now, suppose that we registered this class to have special kwargs for `hydra_zen.builds`:
    monkeypatch.setitem(custom_hydra_zen_builds_args, A, {"zen_exclude": ["a"]})

    schema = _create_schema_for_config(
        config,
        config_file,
        configs_dir=tmp_path,
        repo_root=tmp_path,
        config_store=None,
    )

    mock_hydra_zen_builds.assert_called_once()
    call = mock_hydra_zen_builds.call_args_list[0]
    assert call.args[0] is A
    # We should now expect the `zen_exclude` kwarg to have been passed to `hydra_zen.builds`.
    assert call.kwargs["zen_exclude"] == ["a"]
    file_regression.check(json.dumps(schema, indent=2) + "\n", extension=".json")


class B(A):
    def __init__(self, a: int = 2, b: str = "b"):
        super().__init__(a=a)
        self.b = b


def test_use_custom_builds_kwargs_if_base_in_special_handlers(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
):
    from hydra_auto_schema.customize import custom_hydra_zen_builds_args

    mock_hydra_zen_builds = Mock(spec=hydra_zen.builds, wraps=hydra_zen.builds)
    monkeypatch.setattr(hydra_zen, hydra_zen.builds.__name__, mock_hydra_zen_builds)

    config = {"_target_": get_target_string(B)}
    config_file = tmp_path / "config.yaml"
    omegaconf.OmegaConf.save(config, config_file)

    _schema = _create_schema_for_config(
        config,
        config_file,
        configs_dir=tmp_path,
        repo_root=tmp_path,
        config_store=None,
    )
    mock_hydra_zen_builds.assert_called_once()
    call_without_special_kwargs = mock_hydra_zen_builds.call_args_list[0]
    assert call_without_special_kwargs.args[0] is B
    # The `zen_exclude` kwarg shouldn't be in the call args to start with.
    assert "zen_exclude" not in call_without_special_kwargs

    mock_hydra_zen_builds.reset_mock()

    # Now, suppose that the base class has some special kwargs registered:
    monkeypatch.setitem(custom_hydra_zen_builds_args, A, {"zen_exclude": ["a"]})

    _schema = _create_schema_for_config(
        config,
        config_file,
        configs_dir=tmp_path,
        repo_root=tmp_path,
        config_store=None,
    )

    mock_hydra_zen_builds.assert_called_once()
    call_without_special_kwargs = mock_hydra_zen_builds.call_args_list[0]
    assert call_without_special_kwargs.args[0] is B
    assert call_without_special_kwargs.kwargs["zen_exclude"] == ["a"]



class Foo:
    def __init__(self, bar: str):
        """Description of the `Foo` class.

        Args:
            bar: Description of the `bar` argument.
        """
        self.bar = bar


class Bar(Foo):
    """Docstring of the Bar class.

    Args:
        baz: description of the `baz` argument from the cls docstring instead of the init docstring.
    """

    def __init__(self, bar: str, baz: int):
        # no docstring here.
        super().__init__(bar=bar)
        self.baz = baz



import contextlib
import shutil
import sys
import time
from pathlib import Path
from unittest.mock import Mock

import pytest
from watchdog.events import (
    DirCreatedEvent,  # noqa
    DirDeletedEvent,  # noqa
    DirModifiedEvent,  # noqa
    DirMovedEvent,  # noqa
    FileCreatedEvent,  # noqa
    FileDeletedEvent,  # noqa
    FileModifiedEvent,  # noqa
    FileMovedEvent,  # noqa
)
from watchdog.observers import Observer

from hydra_auto_schema.auto_schema import get_schema_file_path
from hydra_auto_schema.filewatcher import AutoSchemaEventHandler

config_dir = Path(__file__).parent / "configs"


# TODO: Figure out how the calls differ on MacOS.
xfail_on_macos = pytest.mark.xfail(
    sys.platform == "darwin",
    reason="Seems like watchdog uses different events on MacOS.",
    # strict=True,
)


@pytest.fixture
def repo_root(tmp_path: Path):
    repo_root = tmp_path / "repo_root"
    repo_root.mkdir()
    return repo_root


@pytest.fixture
def configs_dir(repo_root: Path):
    test_configs_dir = repo_root / "configs"
    shutil.copytree(config_dir, test_configs_dir)
    shutil.copy(config_dir.parent / "app.py", repo_root)

    # TODO: The structured configs stuff isn't working with the filewatcher or on the CLI for now.
    (test_configs_dir / "__init__.py").unlink()
    (test_configs_dir / "with_structured_default.yaml").unlink()

    return test_configs_dir


@pytest.fixture
def schemas_dir(tmp_path: Path):
    return tmp_path / ".schemas"


@pytest.fixture
def observer():
    """Taken from https://github.com/gorakhargosh/watchdog/blob/master/tests/test_fsevents.py#L39"""
    obs = Observer()
    obs.start()
    try:
        yield obs
    finally:
        obs.stop()
        with contextlib.suppress(RuntimeError):
            obs.join()


@pytest.fixture
def filewatcher(
    repo_root: Path,
    configs_dir: Path,
    schemas_dir: Path,
    observer,
):
    regen_schemas = True
    stop_on_error = True
    quiet = False
    add_headers = True

    # TODO: Do we mock out the calls to hydra stuff? (They are breaking.)

    real_handler = AutoSchemaEventHandler(
        repo_root=repo_root,
        configs_dir=configs_dir,
        schemas_dir=schemas_dir,
        regen_schemas=regen_schemas,
        stop_on_error=stop_on_error,
        quiet=quiet,
        add_headers=add_headers,
    )
    mock_handler = Mock(
        spec=real_handler,
        spec_set=real_handler,
        wraps=real_handler,
    )

    observer.schedule(mock_handler, str(configs_dir), recursive=True)
    yield mock_handler


def _get_all_schema_files(schemas_dir: Path) -> list[Path]:
    return list(schemas_dir.rglob("*.json"))


@xfail_on_macos
def test_on_created(
    filewatcher: AutoSchemaEventHandler, configs_dir: Path, schemas_dir: Path
):
    schemas_before = _get_all_schema_files(schemas_dir)

    new_file = configs_dir / "foobar.yaml"
    assert not new_file.exists()
    schema_file = get_schema_file_path(new_file, schemas_dir)
    assert not schema_file.exists()
    assert schema_file not in schemas_before

    new_file.write_text("foo: bar")

    time.sleep(0.5)

    filewatcher.dispatch.assert_any_call(  # type: ignore
        FileCreatedEvent(
            src_path=str(new_file),
            is_synthetic=False,
        )
    )

    schemas_after = _get_all_schema_files(schemas_dir)
    assert schema_file in schemas_after
    # assert set(schemas_after) == (set(schemas_before) | {schema_file})


@xfail_on_macos
def test_on_modified(
    filewatcher: AutoSchemaEventHandler, configs_dir: Path, schemas_dir: Path
):
    existing_file = configs_dir / "config.yaml"
    assert existing_file.exists()
    schemas_before = _get_all_schema_files(schemas_dir)
    schema_file = get_schema_file_path(existing_file, schemas_dir)
    assert schema_file.exists()
    assert schema_file in schemas_before

    existing_file.write_text(existing_file.read_text() + "\n")
    time.sleep(0.5)

    filewatcher.dispatch.assert_any_call(  # type: ignore
        FileModifiedEvent(
            src_path=str(existing_file),
            is_synthetic=False,
        )
    )

    schemas_after = _get_all_schema_files(schemas_dir)
    assert set(schemas_after) == set(schemas_before)


@xfail_on_macos
def test_on_deleted(
    filewatcher: AutoSchemaEventHandler, configs_dir: Path, schemas_dir: Path
):
    existing_file = configs_dir / "config.yaml"
    assert existing_file.exists()
    schemas_before = _get_all_schema_files(schemas_dir)
    schema_file = get_schema_file_path(existing_file, schemas_dir)
    assert schema_file.exists()
    assert schema_file in schemas_before

    existing_file.unlink()
    time.sleep(0.5)

    filewatcher.dispatch.assert_any_call(  # type: ignore
        FileDeletedEvent(
            src_path=str(existing_file),
            is_synthetic=False,
        )
    )

    schemas_after = _get_all_schema_files(schemas_dir)
    assert set(schemas_after) == set(schemas_before) - {schema_file}


@xfail_on_macos
def test_on_moved(
    filewatcher: AutoSchemaEventHandler, configs_dir: Path, schemas_dir: Path
):
    old_file = configs_dir / "config.yaml"
    new_file = configs_dir / "config2.yaml"
    assert old_file.exists()
    assert not new_file.exists()

    schemas_before = _get_all_schema_files(schemas_dir)
    old_file_schema = get_schema_file_path(old_file, schemas_dir)
    new_file_schema = get_schema_file_path(new_file, schemas_dir)
    assert old_file_schema.exists()
    assert old_file_schema in schemas_before
    assert not new_file_schema.exists()
    assert new_file_schema not in schemas_before  # duh.

    old_file.rename(new_file)
    time.sleep(0.5)

    filewatcher.dispatch.assert_any_call(  # type: ignore
        FileMovedEvent(
            src_path=str(old_file),
            dest_path=str(new_file),
            is_synthetic=False,
        )
    )

    schemas_after = _get_all_schema_files(schemas_dir)
    assert set(schemas_after) == (set(schemas_before) - {old_file_schema}) | {
        new_file_schema
    }



import argparse
import logging
import sys
import time
from pathlib import Path

import rich.logging
from watchdog.observers import Observer

from hydra_auto_schema.auto_schema import add_schemas_to_all_hydra_configs, logger
from hydra_auto_schema.filewatcher import AutoSchemaEventHandler


def main(argv: list[str] | None = None):
    logging.basicConfig(
        level=logging.ERROR,
        # format="%(asctime)s - %(levelname)s - %(message)s",
        format="%(message)s",
        datefmt="[%X]",
        force=True,
        handlers=[
            rich.logging.RichHandler(
                markup=True,
                rich_tracebacks=True,
                tracebacks_width=100,
                tracebacks_show_locals=False,
            )
        ],
    )

    parser = argparse.ArgumentParser()
    parser.add_argument("repo_root", nargs="?", type=Path, default=Path.cwd())
    parser.add_argument(
        "--configs_dir",
        type=Path,
        nargs="?",
        default=None,
        help=(
            "The directory containing hydra config files for which schemas should be generated. "
            "If unset, this will default to the first directory called 'configs' in the repo root."
        ),
    )
    parser.add_argument("--schemas-dir", type=Path, default=Path.cwd() / ".schemas")
    parser.add_argument("--regen-schemas", action=argparse.BooleanOptionalAction)
    parser.add_argument("--stop-on-error", action=argparse.BooleanOptionalAction)
    parser.add_argument("--watch", action=argparse.BooleanOptionalAction)
    parser.add_argument(
        "--add-headers",
        action=argparse.BooleanOptionalAction,
        default=None,
        help=(
            "Always add headers to the yaml config files, instead of the default "
            "behaviour which is to first try to add an entry in the vscode "
            "settings.json file."
        ),
    )
    verbosity_group = parser.add_mutually_exclusive_group()
    verbosity_group.add_argument(
        "-q", "--quiet", dest="quiet", action=argparse.BooleanOptionalAction
    )
    verbosity_group.add_argument(
        "-v", "--verbose", dest="verbose", action="count", default=0
    )

    args = parser.parse_args(argv if argv is not None else sys.argv[1:])

    configs_dir: Path | None = args.configs_dir
    schemas_dir: Path = args.schemas_dir
    repo_root: Path = args.repo_root
    regen_schemas: bool = args.regen_schemas
    stop_on_error: bool = args.stop_on_error
    quiet: bool = args.quiet
    verbose: int = args.verbose
    add_headers: bool = args.add_headers
    watch: bool = args.watch

    repo_root = repo_root.resolve()

    if configs_dir is None:
        configs_dir = next(
            (
                p
                for p in repo_root.rglob("conf*")
                if p.is_dir() and ".venv" not in p.parts
            ),
            None,
        )

    if configs_dir is None:
        raise ValueError(
            f"--configs-dir was not passed, and no hydra configs directory was found in "
            f"the passed repo root: {repo_root}!"
        )
    if not configs_dir.is_absolute():
        configs_dir = repo_root / configs_dir

    configs_dir = configs_dir.resolve()

    if quiet:
        logger.setLevel(logging.NOTSET)
    elif verbose:
        if verbose >= 3:
            logger.setLevel(logging.DEBUG)
        elif verbose == 2:
            logger.setLevel(logging.INFO)
        else:
            assert verbose == 1
            logger.setLevel(logging.WARNING)
    else:
        logger.setLevel(logging.ERROR)
    logger.debug(
        f"{configs_dir=} {schemas_dir=} {repo_root=} {regen_schemas=} {stop_on_error=} {quiet=} {verbose=} {add_headers=} {watch=}"
    )
    # try to find a ConfigStore?

    from hydra.core.config_store import ConfigStore

    config_store = ConfigStore.instance()

    if watch:
        observer = Observer()
        handler = AutoSchemaEventHandler(
            repo_root=repo_root,
            configs_dir=configs_dir,
            schemas_dir=schemas_dir,
            regen_schemas=regen_schemas,
            stop_on_error=stop_on_error,
            quiet=quiet,
            add_headers=add_headers,
            config_store=config_store,
        )
        observer.schedule(handler, str(configs_dir), recursive=True)
        observer.start()
        logger.info("Watching for changes in the config files.")
        try:
            while True:
                time.sleep(1)
        except KeyboardInterrupt:
            observer.stop()
        observer.join()
        return

    add_schemas_to_all_hydra_configs(
        repo_root=repo_root,
        configs_dir=configs_dir,
        schemas_dir=schemas_dir,
        regen_schemas=regen_schemas,
        stop_on_error=stop_on_error,
        quiet=quiet,
        add_headers=add_headers,
        config_store=config_store,
    )
    logger.info("Done updating the schemas for the Hydra config files.")


if __name__ == "__main__":
    main()



import copy
import inspect
from pathlib import Path
from typing import Callable, Mapping, MutableMapping, TypeVar, cast

K = TypeVar("K")
K1 = TypeVar("K1")
K2 = TypeVar("K2")
V = TypeVar("V")
V1 = TypeVar("V1")
V2 = TypeVar("V2")

NestedMapping = MutableMapping[K, V | "NestedMapping[K, V]"]
NestedDict = dict[K, V | "NestedDict[K, V]"]


def get_target_string(some_callable: type | Callable) -> str:
    """Gets the string to use as _target_ in a Hydra config for the given callable.

    >>> get_target_string(int)
    'builtins.int'
    >>> get_target_string(Path)
    'pathlib.Path'
    >>> get_target_string(get_target_string)
    'hydra_auto_schema.utils.get_target_string'
    """
    if inspect.isclass(some_callable):
        return f"{some_callable.__module__}.{some_callable.__qualname__}"
    else:
        name = getattr(
            some_callable, "__qualname__", getattr(some_callable, "__name__")
        )
        return f"{some_callable.__module__}.{name}"


def pretty_path(path: Path) -> str:
    return (
        ("./" + str(path.relative_to(Path.cwd())))
        if path.is_relative_to(Path.cwd())
        else str(path)
    )


def merge_dicts(
    a: NestedMapping[K1, V1],
    b: NestedMapping[K2, V2],
    conflict_handler: Callable[[V1, V2], V1 | V2] | None = None,
    conflict_handlers: dict[str, Callable[[V1, V2], V1 | V2]] | None = None,
    _path: list[str] = [],
) -> NestedMapping[K1 | K2, V1 | V2]:
    """Merge two nested dictionaries.

    NOTE: performs a deep copy of `a` and of values in `b`.

    >>> x = dict(b=1, c=dict(d=2, e=3))
    >>> y = dict(d=3, c=dict(z=2, f=4))
    >>> merge_dicts(x, y)
    {'b': 1, 'c': {'d': 2, 'e': 3, 'z': 2, 'f': 4}, 'd': 3}
    >>> x
    {'b': 1, 'c': {'d': 2, 'e': 3}}
    >>> y
    {'d': 3, 'c': {'z': 2, 'f': 4}}

    If keys are in both dicts and no conflict handler is passed, an error is raised:
    >>> merge_dicts({'a': {'b': 1, 'c': 5}}, {'a': {'b': 2}})
    Traceback (most recent call last):
    ...
    Exception: Conflict at a.b


    in nested dicts conflict, the `conflict_handler` is called with the conflicting values.

    >>> merge_dicts({'a': {'b': 1, 'c': 5}}, {'a': {'b': 2}}, conflict_handler=lambda a, b: a + b)
    {'a': {'b': 3, 'c': 5}}

    Conflict handlers can be passed for specific keys.
    The dict of handlers can map either from the full path or from a key to the handler to use:

    >>> merge_dicts(
    ...     {'a': {'b': 1, 'c': 1, 'd': 12}},
    ...     {'a': {'b': 2, 'c': 2}},
    ...     conflict_handlers={
    ...         'a.b': lambda a, b: a + b,  # this will be used for 'a.b'
    ...         'c': lambda a, b: a - b,    # this will be used for 'c' in any subdict.
    ...     }
    ... )
    {'a': {'b': 3, 'c': -1, 'd': 12}}
    """
    conflict_handlers = conflict_handlers or {}
    out: NestedMapping[K1 | K2, V1 | V2] = copy.deepcopy(a)  # type: ignore
    for key in b:
        b_val: V2 | NestedMapping[K2, V2] = b[key]

        if key not in a:
            out[key] = copy.deepcopy(b_val)  # type: ignore
            continue

        # Type checker doesn't get that `key in a` := `isinstance(key, V1)` yet.
        a_key: K1 = key  # type: ignore
        a_val = a[a_key]

        if isinstance(a_val, Mapping) and isinstance(b_val, Mapping):
            # Type checker doesn't narrow `a_val` or `b_val` to `NestedMapping` yet.
            a_val = cast(NestedMapping[K1, V1], a_val)
            b_val = cast(NestedMapping[K2, V2], b_val)
            out[key] = merge_dicts(
                a_val,
                b_val,
                conflict_handlers={
                    # k: v
                    k.removeprefix(f"{key}."): v
                    for k, v in (conflict_handlers or {}).items()
                },
                conflict_handler=conflict_handler,
                _path=_path + [str(key)],
            )
        elif a_val != b_val:
            # Type checker doesn't narrow `a_val` to `V1` or `b_val` to `V2` yet.
            a_val = cast(V1, a_val)
            b_val = cast(V2, b_val)
            if specific_conflict_handler := conflict_handlers.get(str(key)):
                out[key] = specific_conflict_handler(a_val, b_val)
            elif conflict_handler:
                out[key] = conflict_handler(a_val, b_val)
            else:
                raise Exception("Conflict at " + ".".join(_path + [str(key)]))
    return out



import json
import os
from pathlib import Path

import pytest
import yaml
from hydra.core.config_store import ConfigStore
from pytest_regressions.file_regression import FileRegressionFixture

from hydra_auto_schema.auto_schema import (
    _add_schema_header,
    _create_schema_for_config,
    add_schemas_to_all_hydra_configs,
)

REPO_ROOTDIR = Path.cwd()
IN_GITHUB_CI = "GITHUB_ACTIONS" in os.environ
config_dir = Path(__file__).parent.parent / "tests" / "configs"


@pytest.fixture
def original_datadir():
    return config_dir


test_files = list(config_dir.rglob("*.yaml"))


@pytest.mark.parametrize(
    "config_file",
    [
        pytest.param(
            p,
            marks=pytest.mark.xfail(
                IN_GITHUB_CI,
                reason="TODO: Does not work on the Github CI for some reason!",
            ),
        )
        if "structured" in p.name
        else p
        for p in test_files
    ],
    ids=[f.name for f in test_files],
)
def test_make_schema(config_file: Path, file_regression: FileRegressionFixture):
    """Test that creates a schema for a config file and saves it next to it.

    (in the test folder).
    """
    schema_file = config_file.with_suffix(".json")

    config = yaml.load(config_file.read_text(), yaml.FullLoader)
    if config is None:
        config = {}
    schema = _create_schema_for_config(
        config=config,
        config_file=config_file,
        configs_dir=config_dir,
        repo_root=REPO_ROOTDIR,
        config_store=ConfigStore.instance(),
    )
    _add_schema_header(config_file, schema_path=schema_file)

    file_regression.check(
        json.dumps(schema, indent=2) + "\n", fullpath=schema_file, extension=".json"
    )


def test_warns_when_no_config_files_found(tmp_path: Path):
    with pytest.warns(RuntimeWarning, match="No config files were found"):
        add_schemas_to_all_hydra_configs(
            repo_root=tmp_path, configs_dir=tmp_path, schemas_dir=tmp_path
        )


def test_raises_when_no_config_files_found_and_stop_on_error(tmp_path: Path):
    with pytest.raises(RuntimeError, match="No config files were found"):
        add_schemas_to_all_hydra_configs(
            repo_root=tmp_path,
            configs_dir=tmp_path,
            schemas_dir=tmp_path,
            stop_on_error=True,
        )
